#ifndef __ARRAY_H__
#define __ARRAY_H__

//---------------------------------------------------------------------------
// Array dinamico
//---------------------------------------------------------------------------

#define NoPosition -1;

template <class T> class SArray;
#include "SArray.h"

#include <assert.h>

template <class T> class Array
{
protected:
    T *pArray;			//vettore degli elementi
	
	long nBlockSize;		//dimensione di un blocco di memoria
	long nArraySize;		//dimensione dell'array
	long nAllocatedSize;	//dimensione della memoria allocata (>=nArraySize)
	
	void Resize(long newAllocatedSize);

public:
	T *GetInternalBuffer() { return pArray; }
    
	Array(long initialSize=0, long blockSize=10);
	Array(Array &a);
	
	~Array() { DeleteAll();}
	
	void DeleteAll();

	void SetBlockSize(long newBlockSize, bool updateAllocatedSize=false);
	
	void FreeExtra();

	Array<T>& operator=(Array<T>&);
	Array<T>& operator=(SArray<T>&);
	
	long GetCount() { return nArraySize;}
	
	bool IsEmpty() { return nArraySize==0;}

	void Insert(T t);
	void Delete(T t);
	
	void SetAt(T t, long n);
	T&	 GetAt(long n);
	void InsertAt(T t, long n);
	void DeleteAt(long n);
	
	T& operator[](long n);

	void Reset() {DeleteAll();}
	
	bool Has(T t) {return (Find(t)!=-1);}
	
	long  Find(T t);

	void SetSize(long newSize);

	void ZeroMem() {	memset(pArray, 0, sizeof(T)*nAllocatedSize); }

	void Swap(long i, long j);

	bool OutOfRange(long i) { return !(i>=0 && i<nArraySize); }

#ifdef _DEBUG
	long	GetMemory(bool bAll=false)
	{ return nAllocatedSize*sizeof(T)+(bAll ? sizeof(Array):0); }
	long GetBlockSize() { return nBlockSize; }
#endif //_DEBUG
};

template <class T>
	void Array<T>::Swap(long i, long j)
{
	char szFunction[]="Array<T>::Swap(long i, long j)";

	//controllo la correttezza di i e j

	if(nArraySize==0)
		return;

	if(i==j)
		return;

	if(i<0 || j<0 || i>nArraySize-1 || j>nArraySize-1)
		ASSERT(FALSE);

	T temp;

	temp=pArray[j];
	pArray[j]=pArray[i];
	pArray[i]=temp;
}

//---------------------------------------------------------------------------
// questa funzione viene chiamata solo quando
// il vettore ha effettivamente bisogno di essere 
// ridimensionato, e cio:
// - in seguito ad un inserimento
// - in seguito ad una Resize(newsize) che 
//   richiede un ridimensionamento
// - in seguito ad una FreeExtra che 
//   richiede un ridimensionamento
template <class T>
	void Array<T>::Resize(long newAllocatedSize)
{
	char szFunction[]="Array<T>::Resize(long nNewAllocatedSize)";

	//controllo la correttezza di nNewAllocated size
	if(newAllocatedSize<=0)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_BAD_NEW_SIZE);

	if(nArraySize>newAllocatedSize)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_BAD_NEW_SIZE);

	//se sono uguali non faccio nulla
	if(newAllocatedSize==nAllocatedSize)
		return;

	T *tmpArray;
	tmpArray=new T[newAllocatedSize];

	//usare poi memcpy
	if(pArray!=0)
	{
		long i;
		for(i=0;i<nArraySize;i++)
			tmpArray[i]=pArray[i];

		delete[] pArray;
	}

	nAllocatedSize=newAllocatedSize;
	pArray=tmpArray;
}

//---------------------------------------------------------------------------
template <class T>
	Array<T>::Array(long initialSize, long blockSize)
{
	char szFunction[]="Array<T>::Array(long initialsize, long blocksize)";
	
	if(blockSize<=0 || initialSize<0)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_BAD_NEW_SIZE);

	nBlockSize=blockSize;
	nArraySize=initialSize;

	if(nArraySize!=0)
	{
		nAllocatedSize=nBlockSize*(nArraySize/nBlockSize+1);
		
		if( nArraySize!=0 && (nArraySize%nBlockSize)==0 )
			nAllocatedSize-=nBlockSize;

		pArray=new T[nAllocatedSize];
	}
	else
	{
		pArray=0;
		nAllocatedSize=0;
	}
}

//---------------------------------------------------------------------------
template <class T>
	void Array<T>::DeleteAll()
{
	if(pArray!=0)
	{
		delete[] pArray;
		pArray=0;
	}
	nArraySize=nAllocatedSize=0;
}

//---------------------------------------------------------------------------
template <class T>
	void Array<T>::FreeExtra()
{
	if(IsEmpty())
		return;

	if(nArraySize<=(nAllocatedSize-nBlockSize))
	{
		//calcola la nuova dimensione in base a blocksize
		long nNewAllocatedSize=nBlockSize*(nArraySize/nBlockSize+1);
				
		if( nArraySize!=0 && (nArraySize%nBlockSize)==0 )
			nNewAllocatedSize-=nBlockSize;
		
		Resize(nNewAllocatedSize);
	}
}

//---------------------------------------------------------------------------
template <class T>
	Array<T>::Array(Array<T>& a)
{
	pArray=0;
	nArraySize=nAllocatedSize=0;
	nBlockSize=10;

	if(!a.IsEmpty())
	{
		nBlockSize=a.nBlockSize;
	
		nArraySize=a.GetCount();
		
		nAllocatedSize=nBlockSize*(nArraySize/nBlockSize+1);
	
		if( nArraySize!=0 && (nArraySize%nBlockSize)==0 )
			nAllocatedSize-=nBlockSize;
				
		pArray=new T[nAllocatedSize];

		memcpy(pArray, a.GetInternalBuffer(), nArraySize*sizeof(T));
	}
}

//---------------------------------------------------------------------------
template <class T>
	Array<T>& Array<T>::operator=(Array<T>& a)
{
	DeleteAll();

	if(!a.IsEmpty())
	{
		nArraySize=a.GetCount();
		
		nAllocatedSize=nBlockSize*(nArraySize/nBlockSize+1);
		
		if( nArraySize!=0 && (nArraySize%nBlockSize)==0 )
			nAllocatedSize-=nBlockSize;
				
		pArray=new T[nAllocatedSize];

		/*
		long i;
		for(i=0;i<nArraySize;i++)
			pArray[i]=a[i];
		*/

		memcpy(pArray, a.GetInternalBuffer(), nArraySize*sizeof(T));
	}
	return *this;
}

//---------------------------------------------------------------------------
template <class T>
	Array<T>& Array<T>::operator=(SArray<T>& a)
{
	DeleteAll();

	if(!a.IsEmpty())
	{
		nArraySize=a.GetCount();
		
		nAllocatedSize=nBlockSize*(nArraySize/nBlockSize+1);
		
		if( nArraySize!=0 && (nArraySize%nBlockSize)==0 )
			nAllocatedSize-=nBlockSize;
		
		pArray=new T[nAllocatedSize];

		for(long i=0;i<nArraySize;i++)
			pArray[i]=a[i];
	}
	return *this;
}

//---------------------------------------------------------------------------
template <class T>
	void Array<T>::Insert(T newData)
{
	if(nArraySize==nAllocatedSize)
	{
		//alloco altri blocksize elementi
		Resize(nAllocatedSize+nBlockSize);
	}

	pArray[nArraySize]=newData;
	nArraySize++;
}

//---------------------------------------------------------------------------
template <class T>
	void Array<T>::Delete(T delData)
{
	long i,j;

	if(IsEmpty())
		return;

	for(i=0;i<nArraySize && pArray[i]!=delData;i++)
		;

	if(i!=nArraySize)
	{
		if(nArraySize==1)
			DeleteAll();
		else
		{
			for(j=i+1;j<nArraySize;j++)
				pArray[j-1]=pArray[j];
			nArraySize--;
		}
	}
	else
	{
		//messaggio errore-non trovato
		return;
	}
}

//---------------------------------------------------------------------------
template <class T>
	void Array<T>::SetAt(T newData, long n)
{
	char szFunction[]="Array<T>::SetAt(T newitem, long n)";

	if( IsEmpty() || n<0 || n>= nArraySize)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);

	pArray[n]=newData;	
}

//---------------------------------------------------------------------------
template <class T>
	T& Array<T>::GetAt(long n)
{
	char szFunction[]="Array<T>::GetAt(long n) ";

	if( IsEmpty() || n<0 || n>= nArraySize)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);

	return pArray[n];
}

//---------------------------------------------------------------------------
template <class T>
	void Array<T>::InsertAt(T newData, long n)
{
	char szFunction[]="Array<T>::InsertAt(T newitem, long n)";

	if(n==0 && IsEmpty())
		Insert(newData);
	else
	{
		if( IsEmpty() || n<0 || n>= nArraySize)
			ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);

		if(nArraySize==nAllocatedSize)
		{
			Resize(nAllocatedSize+nBlockSize);
		}

		//sposta in avanti di uno gli
		//elementi con indice superiore a n
		for(long i=nArraySize-1;i>=n;i--)
			pArray[i+1]=pArray[i];

		pArray[n]=newData;
		nArraySize++;
	}
}

//---------------------------------------------------------------------------
template <class T>
	void Array<T>::DeleteAt(long n)
{
	char szFunction[]="Array<T>::DeleteAt(long n)";

	if( IsEmpty() || n<0 || n>= nArraySize)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);

	for(long i=n+1;i<nArraySize;i++)
		pArray[i-1]=pArray[i];
	nArraySize--;
}

//---------------------------------------------------------------------------
template <class T>
	T& Array<T>::operator[](long n)
{
	char szFunction[]="Array<T>::GetAt(long n) ";

	if( IsEmpty() || n<0 || n>= nArraySize)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);
		
	return pArray[n];
}

//---------------------------------------------------------------------------
template <class T>
	long Array<T>::Find(T findData)
{
	if(IsEmpty())
		return -1;

	long i;
	for(i=0;i<nArraySize && pArray[i]!=findData;i++)
		;
	if(i==nArraySize)
		return -1;

	return i;
}

//---------------------------------------------------------------------------
template <class T>
	void Array<T>::SetSize(long newSize)
{
	char szFunction[]="Array<T>::SetSize(long newsize)";

	if(newSize<0)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_BAD_NEW_SIZE);

	if(newSize==nArraySize)
		return;

	if(newSize==0)
		DeleteAll();
	else
	{
		if(newSize<nArraySize)
		{
			nArraySize=newSize;
			//qui devo fare l'assegnazione prima perch
			//devo conoscere la nuova dimensione dell'array
			//per poterlo accorciare
			FreeExtra();
		}
		else if(newSize>nAllocatedSize)
		{
			
			long nNewAllocatedSize=nBlockSize*(newSize/nBlockSize+1);
		
			if( newSize!=0 && (newSize%nBlockSize)==0 )
				nNewAllocatedSize-=nBlockSize;
					
			Resize(nNewAllocatedSize);
		}
	}
	
	//in ogni caso devo fare l'assegnazione
	nArraySize=newSize;
}

//---------------------------------------------------------------------------
template <class T>
	void Array<T>::SetBlockSize(long newBlockSize, bool bUpdateAllocatedSize)
{
	if(newBlockSize<=0)
		return;

	if(newBlockSize==nBlockSize)
		return;

	if(bUpdateAllocatedSize)
	{
		//qui devo sincronizzare nAllocatedSize
		long nNewAllocatedSize=newBlockSize*(nArraySize/newBlockSize+1);
		
		if( nArraySize!=0 && (nArraySize%newBlockSize)==0 )
				nNewAllocatedSize-=newBlockSize;
		
		Resize(nNewAllocatedSize);
	}
	
	nBlockSize=newBlockSize;

	//attenzione: nel secondo caso occorre che si faccia
	//attenzione a non chiamare la freeextra perch 
	//potrebbe dare risultato errato
	
}

#endif //__ARRAY_H__