#ifndef _SARRAY_H_
#define _SARRAY_H_

//---------------------------------------------------------------------------
// Array dinamico - ottimizza l'inserimento di un gran numero di elementi
// specialmente in coda
//---------------------------------------------------------------------------

#define NoPosition -1;

template <class T> class Array;
#include "Array.h"

#include <assert.h>

template <class T> class SArray
{
protected:
    T **pArray;			//vettore dei blocchi
	
	long nBlockSize;		//dimensione di un blocco
	long nArraySize;		//dimensione dell'array
	long nAllocatedSize;	//dimensione della memoria allocata (>=nArraySize)
	long nBlockCount;	//numero di blocchi

	void Resize(long newAllocatedSize, bool blockSizeChanged=false, long newBlockSize=0);

public:

	T **GetInternalBuffer() {return pArray;}
    
	SArray(long initialSize=0, long blockSize=10);
	SArray(SArray &a);
	
	~SArray() { DeleteAll();}
	
	void DeleteAll();
	
	void SetBlockSize(long newBlockSize, bool updateAllocatedSize=true);
	
	void FreeExtra();

	SArray<T>& operator=(Array<T>&);
	SArray<T>& operator=(SArray<T>&);
	
	long GetCount() { return nArraySize;}

	bool IsEmpty() { return nArraySize==0;}

	void Insert(T t);
	void Delete(T t);
	
	void SetAt(T t, long n);
	T&	 GetAt(long n);
	void InsertAt(T t, long n);
	void DeleteAt(long n);
	
	T& operator[](long n);

	void Reset() {DeleteAll();}
	
	bool Has(T t) {return (Find(t)!=-1);}
	
	long  Find(T t);
	
	void SetSize(long newSize);

	void ZeroMem();

	void Swap(long i, long j);

	bool OutOfRange(long i) { return !(i>=0 && i<nArraySize); }

#ifdef _DEBUG
	long	GetMemory(bool bAll=false)
	{ return nAllocatedSize*sizeof(T)+nBlockCount*sizeof(T *) +(bAll ? sizeof(SArray):0); }
	long GetBlockSize() { return nBlockSize; }
	long GetBlockCount() { return nBlockCount; }
#endif //_DEBUG

};

#define BLOCK(i, b) ((i)/(b))
#define ITEM(i, b)  ((i)%(b))

template <class T>
	void SArray<T>::Swap(long i, long j)
{
	char szFunction[]="SArray<T>::Swap(long i, long j)";

	//controllo la correttezza di i e j

	if(nArraySize==0)
		return;

	if(i==j)
		return;

	if(i<0 || j<0 || i>nArraySize-1 || j>nArraySize-1)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);

	T temp;

	temp=pArray[BLOCK(j, nBlockSize)][ITEM(j, nBlockSize)];
	pArray[BLOCK(j, nBlockSize)][ITEM(j, nBlockSize)]=pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)];
	pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)]=temp;
}


template <class T>
	void SArray<T>::ZeroMem()
{
	for(long i=0; i<nBlockCount; i++)
		memset(pArray[i], 0, sizeof(T)*nBlockSize);
}

//---------------------------------------------------------------------------
// questa funzione viene chiamata solo quando
// il vettore ha effettivamente bisogno di essere 
// ridimensionato, e cio:
// - in seguito ad un inserimento
// - in seguito ad una Resize(newsize) che 
//   richiede un ridimensionamento
// - in seguito ad una FreeExtra che 
//   richiede un ridimensionamento
template <class T>
	void SArray<T>::Resize(long newAllocatedSize, bool blockSizeChanged, long newBlockSize)
{
	char szFunction[]="SArray<T>::Resize(long nNewAllocatedSize)";

	//controllo la correttezza di nNewAllocated size
	if(newAllocatedSize<=0)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_BAD_NEW_SIZE);

	if(nArraySize>newAllocatedSize)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_BAD_NEW_SIZE);

	long nNewBlockCount;
	T **tmpArray;
	
	long i;

	if(blockSizeChanged)
	{
		//se  cambiata la dimensione del blocco allora
		//devo rivoluzionare tutto
		
		//nuovo numero di blocchi necessario
		nNewBlockCount=(newAllocatedSize/newBlockSize)+1;
		
		if( newAllocatedSize!=0 && (newAllocatedSize%newBlockSize)==0 )
			nNewBlockCount-=1;
				
		//alloca i puntatori ai blocchi
		tmpArray=new T*[nNewBlockCount];
		
		//quanto fatto dopo si pu ottimizzare
		//creando memoria e cancellandola a mano a mano

		//alloco lo spazio
		for(i=0; i<nNewBlockCount; i++)
			tmpArray[i]=new T[newBlockSize];

		if(pArray!=0)
		{
			//ricopio i dati
			for(i=0; i<nArraySize; i++)
				tmpArray[BLOCK(i, newBlockSize)][ITEM(i, newBlockSize)]=pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)];
			
			//cancello i vecchi dati
			for(i=0; i<nBlockCount; i++)
				delete[] pArray[i];
			
			delete[] pArray;
		}

		nBlockSize=newBlockSize;
	}
	else
	{
		//se sono uguali 
		if(newAllocatedSize==nAllocatedSize)
			return;
		else if(newAllocatedSize>nAllocatedSize)
		{
			//se  maggiore

			//nuovo numero di blocchi necessario
			nNewBlockCount=(newAllocatedSize/nBlockSize)+1;			

			if( newAllocatedSize!=0 && (newAllocatedSize%nBlockSize)==0 )
				nNewBlockCount-=1;
				
			//alloca i puntatori ai blocchi
			tmpArray=new T*[nNewBlockCount];
			
			if(pArray!=0)
			{
				//copio gli array esistenti
				for(i=0;i<nBlockCount;i++)
					tmpArray[i]=pArray[i];
				
				//cancello il vecchio array
				delete[] pArray;
			}

			//alloco nuovo spazio
			for(i=nBlockCount;i<nNewBlockCount;i++)
				tmpArray[i]=new T[nBlockSize];
		}
		else
		{
			//se  minore
		
			//nuovo numero di blocchi necessario
			nNewBlockCount=(newAllocatedSize/nBlockSize)+1;			
			
			if( newAllocatedSize!=0 && (newAllocatedSize%nBlockSize)==0 )
				nNewBlockCount-=1;
		
			//alloca i puntatori ai blocchi
			tmpArray=new T*[nNewBlockCount];
			
			//qui pArray deve essere per forza != 0

			//copio gli array esistenti
			for(i=0;i<nNewBlockCount;i++)
				tmpArray[i]=pArray[i];
			
			//dealloco lo spazio in eccesso
			for(i=nNewBlockCount;i<nBlockCount;i++)
				delete[] pArray[i];
			
			//cancello il vecchio array
			delete[] pArray;
		}
	}
	
	//aggiornamento di nBlockCount
	nBlockCount=nNewBlockCount;
	nAllocatedSize=newAllocatedSize;
	pArray=tmpArray;
}

//---------------------------------------------------------------------------
template <class T>
	SArray<T>::SArray(long initialSize, long blockSize)
{
	char szFunction[]="SArray<T>::SArray(long initialsize, long blocksize)";
	
	if(blockSize<=0 || initialSize<0)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_BAD_NEW_SIZE);

	nBlockSize=blockSize;
	nArraySize=initialSize;
	
	if(nArraySize!=0)
	{
		nBlockCount=(nArraySize/nBlockSize)+1;

		if( nArraySize!=0 && (nArraySize%nBlockSize)==0 )
			nBlockCount-=1;

		nAllocatedSize=nBlockCount*nBlockSize;
		pArray=new T*[nBlockCount];
		long i;
		for(i=0;i<nBlockCount;i++)
			pArray[i]=new T[nBlockSize];	
	}
	else
	{
		nBlockCount=0;
		nAllocatedSize=0;
		pArray=0;
	}
}

//---------------------------------------------------------------------------
template <class T>
	void SArray<T>::DeleteAll()
{
	if(pArray!=0)
	{
		long i;
		for(i=0;i<nBlockCount;i++)
			delete[] pArray[i];	
		delete[] pArray;
		pArray=0;
	}
	nArraySize=nAllocatedSize=nBlockCount=0;
}

//---------------------------------------------------------------------------
template <class T>
	void SArray<T>::FreeExtra()
{
	if(IsEmpty())
		return;

	if(nArraySize<=(nAllocatedSize-nBlockSize))
	{
		//calcola la nuova dimensione in base a blocksize
		long nNewBlockCount=(nArraySize/nBlockSize)+1;

		if( nArraySize!=0 && (nArraySize%nBlockSize)==0 )
			nNewBlockCount-=1;
		
		long nNewAllocatedSize=nNewBlockCount*nBlockSize;		
		
		Resize(nNewAllocatedSize);
	}
}

//---------------------------------------------------------------------------
template <class T>
	SArray<T>::SArray(SArray<T>& a)
{
	pArray=0;
	nArraySize=nAllocatedSize=0;
	nBlockSize=10;

	if(!a.IsEmpty())
	{
		nBlockSize=a.nBlockSize;

		nArraySize=a.GetCount();
		
		nBlockCount=(nArraySize/nBlockSize)+1;

		if( nArraySize!=0 && (nArraySize%nBlockSize)==0 )
			nBlockCount-=1;
		
		nAllocatedSize=nBlockCount*nBlockSize;

		long i;
		pArray=new T*[nBlockCount];
		for(i=0;i<nBlockCount;i++)
			pArray[i]=new T[nBlockSize];

		//memcpy(pArray, a.GetInternalBuffer(), nArraySize*sizeof(T));

		//copia ad uno ad uno gli elementi del
		//vecchio vettore
		//Nota: i blocchi possono essere diversi!
		for(i=0;i<nArraySize;i++)
			pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)]=a[i];    	

		//memcpy(pArray, a.GetInternalBuffer(), nArraySize*sizeof(T));
	}
}

//---------------------------------------------------------------------------
template <class T>
	SArray<T>& SArray<T>::operator=(Array<T>& a)
{
	DeleteAll();

	if(!a.IsEmpty())
	{
		nArraySize=a.GetCount();
		
		nBlockCount=(nArraySize/nBlockSize)+1;

		if( nArraySize!=0 && (nArraySize%nBlockSize)==0 )
			nBlockCount-=1;
		
		nAllocatedSize=nBlockCount*nBlockSize;

		long i;
		pArray=new T*[nBlockCount];
		for(i=0;i<nBlockCount;i++)
			pArray[i]=new T[nBlockSize];

		//copia ad uno ad uno gli elementi del
		//vecchio vettore
		//Nota: i blocchi possono essere diversi!
		for(i=0;i<nArraySize;i++)
			pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)]=a[i];    	
	}
	return *this;
}

//---------------------------------------------------------------------------
template <class T>
	SArray<T>& SArray<T>::operator=(SArray<T>& a)
{
	DeleteAll();

	if(!a.IsEmpty())
	{
		nArraySize=a.GetCount();
		
		nBlockCount=(nArraySize/nBlockSize)+1;

		if( nArraySize!=0 && (nArraySize%nBlockSize)==0 )
			nBlockCount-=1;
		
		nAllocatedSize=nBlockCount*nBlockSize;

		long i;
		pArray=new T*[nBlockCount];
		for(i=0;i<nBlockCount;i++)
			pArray[i]=new T[nBlockSize];

		//copia ad uno ad uno gli elementi del
		//vecchio vettore
		//Nota: i blocchi possono essere diversi!
		
		for(i=0;i<nArraySize;i++)
			pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)]=a[i];    	
	}
	return *this;
}

//---------------------------------------------------------------------------
template <class T>
	void SArray<T>::Insert(T newData)
{
	if(nArraySize==nAllocatedSize)
	{
		//alloco altri blocksize elementi
		Resize(nAllocatedSize+nBlockSize);
	}
	
	pArray[BLOCK(nArraySize, nBlockSize)][ITEM(nArraySize, nBlockSize)]=newData;
	
	nArraySize++;
}

//---------------------------------------------------------------------------
template <class T>
	void SArray<T>::Delete(T delData)
{
	long i,j;
	
	if(IsEmpty())
		return;
	
	for(i=0;i<nArraySize && pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)]!=delData;i++)
		;
		
	if(i!=nArraySize)
	{
		if(nArraySize==1)
			DeleteAll();
		else
		{
			//T temp;

			//qui devo shiftare tutto indietro di uno!
			for(j=i+1;j<nArraySize;j++)
			{
				pArray[BLOCK(j-1, nBlockSize)][ITEM(j-1, nBlockSize)]=pArray[BLOCK(j, nBlockSize)][ITEM(j, nBlockSize)];
				//temp=pArray[BLOCK(l, nBlockSize)][ITEM(l, nBlockSize)];
				//pArray[BLOCK(l-1, nBlockSize)][ITEM(l-1, nBlockSize)]=temp;				
			}
			nArraySize--;
		}
	}
	else
	{
		//messaggio errore
		return;
	}
}

//---------------------------------------------------------------------------
template <class T>
	void SArray<T>::SetAt(T newData, long n)
{
	char szFunction[]="SArray<T>::SetAt(T newitem, long n)";

	if( IsEmpty() || n<0 || n>= nArraySize)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);
			
	pArray[BLOCK(n, nBlockSize)][ITEM(n, nBlockSize)]=newData;
}

//---------------------------------------------------------------------------
template <class T>
	T& SArray<T>::GetAt(long n) 
{
	char szFunction[]="SArray<T>::GetAt(long n) ";

	if( IsEmpty() || n<0 || n>= nArraySize)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);
	
	return pArray[BLOCK(n, nBlockSize)][ITEM(n, nBlockSize)];
}

//---------------------------------------------------------------------------
template <class T>
	void SArray<T>::InsertAt(T newData, long n)
{
	char szFunction[]="SArray<T>::InsertAt(T newitem, long n)";

	if(n==0 && IsEmpty())
		Insert(newData);
	else
	{
		if( IsEmpty() || n<0 || n>= nArraySize)
			ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);
	
		if(nArraySize==nAllocatedSize)
		{
			Resize(nAllocatedSize+nBlockSize);
		}
		
		//T temp;
		//qui devo shiftare tutto avanti di uno!
		//long i;

		for(long i=nArraySize-1;i>=n;i--)
		{
			pArray[BLOCK(i+1, nBlockSize)][ITEM(i+1, nBlockSize)]=pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)];
			
			//temp=pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)];
			//pArray[BLOCK(i+1, nBlockSize)][ITEM(i+1, nBlockSize)]=temp;				
		}

		pArray[BLOCK(n, nBlockSize)][ITEM(n, nBlockSize)]=newData;
		
		nArraySize++;
	}
}

//---------------------------------------------------------------------------
template <class T>
	void SArray<T>::DeleteAt(long n)
{
	char szFunction[]="SArray<T>::DeleteAt(long n)";

	
	if( IsEmpty() || n<0 || n>= nArraySize)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);
	
	//T temp;
	//qui devo shiftare tutto indietro di uno!
	long i;
	for(i=n+1;i<nArraySize;i++)
	{
		pArray[BLOCK(i-1, nBlockSize)][ITEM(i-1, nBlockSize)]=pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)];

		//temp=pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)];
		//pArray[BLOCK(i-1, nBlockSize)][ITEM(i-1, nBlockSize)]=temp;				
	}
	
	nArraySize--;
}

//---------------------------------------------------------------------------
template <class T>
	T& SArray<T>::operator[](long n) 
{
	char szFunction[]="SArray<T>::GetAt(long n) ";

	if( IsEmpty() || n<0 || n>= nArraySize)
		//ASSERT(FALSE);
		assert(false); //throw MException(szFunction, IDS_ERR_ARRAY_INDEX_OUT_OF_RANGE);
	
	return pArray[BLOCK(n, nBlockSize)][ITEM(n, nBlockSize)];
}

//---------------------------------------------------------------------------
template <class T>
	long SArray<T>::Find(T findData)
{
	if(IsEmpty())
		return -1;

	long i;
	for(i=0;i<nArraySize && pArray[BLOCK(i, nBlockSize)][ITEM(i, nBlockSize)]!=findData;i++)
		;
	
	if(i==nArraySize)
		return -1;

	return i;
}

//---------------------------------------------------------------------------
template <class T>
	void SArray<T>::SetSize(long newSize)
{
	char szFunction[]="SArray<T>::SetSize(long newsize)";

	if(newSize<0)
		ASSERT(FALSE); //throw MException(szFunction, IDS_ERR_ARRAY_BAD_NEW_SIZE);

	if(newSize==nArraySize)
		return;

	if(newSize==0)
		DeleteAll();
	else
	{
		if(newSize<nArraySize)
		{
			nArraySize=newSize;
			//qui devo fare l'assegnazione prima perch
			//devo conoscre la nuova dimensione dell'array
			//per poterlo accorciare
			FreeExtra();
		}
		else if(newSize>nAllocatedSize)
		{
			
			long nNewAllocatedSize=nBlockSize*(newSize/nBlockSize+1);
		
			if( newSize!=0 && (newSize%nBlockSize)==0 )
				nNewAllocatedSize-=nBlockSize;
		
			Resize(nNewAllocatedSize);
		}
	}

	//in ogni caso devo fare l'assegnazione
	nArraySize=newSize;
}

//---------------------------------------------------------------------------
//in questo caso cambiando il blocco occorre 
//ricopiare tutto (per la dimensione fa ricorso
//a nAllocatedSize e non a nArraySize)
template <class T>
	void SArray<T>::SetBlockSize(long newBlockSize, bool updateAllocatedSize)
{
	if(newBlockSize<=0)
		return;

	if(newBlockSize==nBlockSize)
		return;

	//qui devo sincronizzare nAllocatedSize

	long nNewAllocatedSize=newBlockSize*(nArraySize/newBlockSize+1);
		
	if( nArraySize!=0 && (nArraySize%newBlockSize)==0 )
			nNewAllocatedSize-=newBlockSize;
		
	// cambiato il blocksize, devo riaggiustare tutto
	Resize(nNewAllocatedSize, true, newBlockSize);
	
	nBlockSize=newBlockSize;
}	


#endif //_SARRAY_H_